#!/usr/bin/python3
#
# Shell handling functions.
#
# Copyright (c) 2018-2019 Xevo Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import os
import shutil
import subprocess


from wst import (
    dry_run,
    log,
    log_cmd
)


def mkdir(path):
    '''Makes a directory.'''
    log('making directory %s' % path)
    if not dry_run():
        os.mkdir(path)


def symlink(dest, src):
    '''Makes a symlink from src to dest.'''
    log('making symlink %s --> %s' % (src, dest))
    if not dry_run():
        os.symlink(dest, src)


def rename(old, new):
    '''Renames the given path.'''
    log('renaming %s --> %s' % (old, new))
    if not dry_run():
        os.rename(old, new)


def remove(path, fail_ok=False):
    '''Removes the given path.'''
    log('removing %s' % path)
    if not dry_run():
        try:
            os.unlink(path)
        except FileNotFoundError:
            if not fail_ok:
                raise


def rmtree(path, fail_ok=False):
    '''Removes the given file or directory, recursively.'''
    log('recursively removing %s' % path)
    if not dry_run():
        try:
            shutil.rmtree(path)
        except FileNotFoundError:
            if not fail_ok:
                raise


def get_shell():
    '''Gets the current default shell.'''
    try:
        return os.environ['SHELL']
    except KeyError:
        # Default, should exist on all systems.
        return '/bin/sh'


def call(cmd, **kwargs):
    '''Calls a given command with the given environment, swallowing the
    output.'''
    log_cmd(cmd)
    if not dry_run():
        subprocess.check_call(cmd, **kwargs)


def call_output(cmd, env=None, text=True, override=False):
    '''Calls a given command with the given environment, returning the
    output. Note that we assume the output is UTF-8.'''
    log_cmd(cmd)
    if (not dry_run()) or override:
        out = subprocess.check_output(cmd, env=env)
        if text:
            out = out.decode('utf-8')
        return out


def call_git(repo, subcmd):
    '''Executes a git command in a given repository.'''
    return call_output(('git', '-C', repo) + subcmd, text=False)


def call_noexcept(op, cmd, **kwargs):
    '''Calls a given command, swallowing output, and returns False if the
    command failed. Normally, we would just crash with an exeception. This
    should be used for commands that are allowed to fail in a normal case.'''
    try:
        call(cmd, **kwargs)
    except subprocess.CalledProcessError:
        return False
    return True


def call_configure(cmd, **kwargs):
    '''Executes a configure command.'''
    return call_noexcept('configure', cmd, **kwargs)


def call_build(cmd, **kwargs):
    '''Executes a build command.'''
    return call_noexcept('build', cmd, **kwargs)


def call_clean(cmd, **kwargs):
    '''Executes a clean command.'''
    return call_noexcept('clean', cmd, **kwargs)



#!/usr/bin/python3
#
# Remove action implementation.
#
# Copyright (c) 2018-2019 Xevo Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import os

from wst import WSError
from wst.cmd import Command
from wst.conf import (
    get_default_ws_link,
    get_ws_dir
)
from wst.shell import (
    remove,
    rmtree,
    symlink
)


class Remove(Command):
    '''The remove command.'''
    @classmethod
    def args(cls, parser):
        '''Populates the argument parser for the remove command.'''
        parser.add_argument(
            metavar='workspace',
            dest='remove_ws',
            action='store',
            help='Workspace to remove')
        parser.add_argument(
            '-d', '--default',
            action='store',
            default=None,
            help='New default workspace')

    @classmethod
    def do(cls, _, args):
        '''Executes the remove command.'''
        ws_dir = get_ws_dir(args.root, args.remove_ws)
        if not os.path.exists(ws_dir):
            raise WSError('workspace %s does not exist' % args.remove_ws)

        if args.default is not None:
            default_ws_dir = get_ws_dir(args.root, args.default)
            if not os.path.exists(default_ws_dir):
                raise WSError('workspace %s does not exist' % args.default)

        default_link = get_default_ws_link(args.root)
        is_default = (os.readlink(default_link) == args.remove_ws)
        if is_default:
            # If the deleted workspace is the default, force the user to choose
            # a new one.
            if args.default is None:
                raise WSError('trying to remove the default workspace; must '
                              'specify a new default via -d/--default')
        elif args.default:
            raise WSError('-d/--default is not applicable unless you are '
                          'removing the default workspace')

        # We are good to go.
        rmtree(ws_dir)
        if is_default:
            remove(default_link)
            symlink(args.default, default_link)



#!/usr/bin/python3
#
# List action implementation.
#
# Copyright (c) 2018-2019 Xevo Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import os

from wst.cmd import Command
from wst.conf import (
    get_default_ws_name,
    get_manifest_link_name,
    parse_manifest
)


class List(Command):
    '''The list command.'''
    @classmethod
    def args(cls, parser):
        '''Populates the argument parser for the list subcmd.'''
        parser.add_argument(
            '-w', '--workspaces',
            action='store_true',
            dest='list_workspaces',
            default=False,
            help='List workspaces instead of projects')

    @classmethod
    def do(cls, _, args):
        '''Executes the list subcmd.'''
        if args.list_workspaces:
            dirs = os.listdir(args.root)
            for ws in dirs:
                if ws == get_default_ws_name():
                    continue
                if ws == get_manifest_link_name():
                    continue
                print(ws)
        else:
            d = parse_manifest(args.root)
            for proj in d:
                print(proj)



#!/usr/bin/python3
#
# Base class for builders.
#
# Copyright (c) 2018-2019 Xevo Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#


class Builder(object):

    '''A builder, representing the interaction with an underlying build
    system, such as CMake or meson. Since this is really an interface and not a
    class, all methods on it should be class methods.'''
    @classmethod
    def env(cls,
            proj,
            prefix,
            build_dir,
            source_dir,
            env,
            build_type,
            builder_args):
        raise NotImplementedError

    @classmethod
    def conf(cls,
             proj,
             prefix,
             build_dir,
             source_dir,
             env,
             build_type,
             builder_args):
        raise NotImplementedError

    @classmethod
    def build(cls,
              proj,
              source_dir,
              build_dir,
              env,
              targets,
              builder_args,
              args):
        raise NotImplementedError

    @classmethod
    def clean(cls, proj, build_dir, env, builder_args):
        raise NotImplementedError



#!/usr/bin/python3
#
# General functions usable by the entire module.
#
# Copyright (c) 2018-2019 Xevo Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#


import logging


# Constants used by multiple modules. Normally we would put these in conf, but
# that can result in a circular import because conf imports builder.* and
# builder.* imports conf.
BUILD_TYPES = ('debug', 'release')
DEFAULT_TARGETS = ('install'),


class WSError(Exception):
    def __init__(self, msg):
        super().__init__(msg)


_DRY_RUN = False
def dry_run():  # noqa: E302
    '''Retrieves the current dry run setting.'''
    global _DRY_RUN
    return _DRY_RUN


def set_dry_run():
    '''Sets whether or not to use dry run mode.'''
    global _DRY_RUN
    _DRY_RUN = True


def log(msg, level=logging.DEBUG):
    '''Logs a message.'''
    logging.log(level, msg)


def log_cmd(cmd, level=logging.DEBUG):
    '''Logs a given command being run.'''
    log(' '.join(cmd), level=level)



#!/usr/bin/python3
#
# Default action implementation.
#
# Copyright (c) 2018-2019 Xevo Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import os

from wst import WSError
from wst.cmd import Command
from wst.conf import (
    get_default_ws_link,
    get_ws_dir
)
from wst.shell import (
    remove,
    symlink
)


class Default(Command):
    '''The default command.'''
    @classmethod
    def args(cls, parser):
        '''Populates the argument parser for the default command.'''
        parser.add_argument(
            metavar='workspace',
            dest='default_ws',
            action='store',
            default=None,
            nargs='?',
            help='Workspace to make the default')

    @classmethod
    def do(cls, _, args):
        '''Executes the default command.'''
        link = get_default_ws_link(args.root)
        if args.default_ws is None:
            # Just report what the default currently is.
            dest = os.path.basename(os.path.realpath(link))
            print(dest)
            return

        ws_dir = get_ws_dir(args.root, args.default_ws)

        remove(link)
        if not os.path.exists(ws_dir):
            raise WSError('Cannot make non-existent workspace %s the default' %
                          args.default_ws)

        symlink(args.default_ws, link)



#!/usr/bin/python3
#
# Clean action implementation.
#
# Copyright (c) 2018-2019 Xevo Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import errno
import os

from wst import (
    dry_run,
    log,
    WSError
)
from wst.cmd import Command
from wst.conf import (
    get_build_dir,
    get_build_env,
    get_builder,
    get_install_dir,
    get_source_dir,
    get_ws_config,
    invalidate_checksum,
    parse_manifest
)
from wst.shell import rmtree


def _force_clean(ws, proj):
    '''Performs a force-clean of a project, removing all files instead of
    politely calling the clean function of the underlying build system.'''
    build_dir = get_build_dir(ws, proj)
    log('removing %s' % build_dir)
    if dry_run():
        return
    try:
        rmtree(build_dir)
    except OSError as e:
        if e.errno == errno.ENOENT:
            log('%s already removed' % build_dir)
        else:
            raise

    config = get_ws_config(ws)
    config['projects'][proj]['taint'] = False


def _polite_clean(root, ws, proj, d):
    '''Performs a polite-clean of a project, calling the underlying build
    system of a project and asking it to clean itself.'''
    builder = get_builder(d, proj)
    build_dir = get_build_dir(ws, proj)
    if not os.path.exists(build_dir):
        return

    build_env = get_build_env(ws, d, proj)
    prefix = get_install_dir(ws, proj)
    source_dir = get_source_dir(root, d, proj)
    builder.clean(
        proj,
        prefix,
        source_dir,
        build_dir,
        build_env,
        d[proj]['builder-args'])


def clean(root, ws, proj, d, force):
    '''Cleans a project, forcefully or not.'''
    invalidate_checksum(ws, proj)

    if force:
        _force_clean(ws, proj)
    else:
        _polite_clean(root, ws, proj, d)


class Clean(Command):
    '''The clean command.'''
    @classmethod
    def args(cls, parser):
        '''Populates the argument parser for the clean command.'''
        parser.add_argument(
            'projects',
            action='store',
            nargs='*',
            help='Clean project(s)')
        parser.add_argument(
            '-f', '--force',
            action='store_true',
            default=False,
            help='Force-clean (remove the build directory)')

    @classmethod
    def do(cls, ws, args):
        '''Executes the clean command.'''
        # Validate.
        d = parse_manifest(args.root)
        for project in args.projects:
            if project not in d:
                raise WSError('unknown project %s' % project)

        if len(args.projects) == 0:
            projects = d.keys()
        else:
            projects = args.projects

        for project in projects:
            clean(args.root, ws, project, d, args.force)



#!/usr/bin/python3
#
# Functions for handling internal ws configuration.
#
# Copyright (c) 2018-2019 Xevo Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import collections
import copy
import errno
import hashlib
import logging
import os
import yaml

from wst import (
    DEFAULT_TARGETS,
    dry_run,
    log,
    WSError
)
from wst.shell import (
    call_git,
    call_output,
    remove,
    rmtree
)

from wst.builder.cmake import CMakeBuilder  # noqa: E402
from wst.builder.meson import MesonBuilder  # noqa: E402
from wst.builder.setuptools import SetuptoolsBuilder  # noqa: E402


_REQUIRED_KEYS = {'build'}
_OPTIONAL_KEYS = {'deps', 'env', 'args', 'builder-args', 'targets'}
_ALL_KEYS = _REQUIRED_KEYS.union(_OPTIONAL_KEYS)
def parse_yaml(root, manifest):  # noqa: E302
    '''Parses the given manifest for YAML and syntax correctness, or bails if
    something went wrong.'''
    try:
        with open(manifest, 'r') as f:
            d = yaml.safe_load(f)
    except IOError:
        raise WSError('ws manifest %s not found' % manifest)

    try:
        includes = d['include']
    except KeyError:
        includes = ()
        d['include'] = includes
    else:
        if not isinstance(includes, list):
            raise WSError('"include" key %s in %s is not a list'
                          % (includes, manifest))
        if len(includes) == 0:
            raise WSError('"include" key in %s is an empty list!' % manifest)

    try:
        search_paths = d['search-path']
    except KeyError:
        search_paths = ()
        d['search-path'] = search_paths
    else:
        if not isinstance(search_paths, list):
            raise WSError('"search-path" key %s in %s is not a list'
                          % (search_paths, manifest))
        if len(search_paths) == 0:
            raise WSError('include in %s is an empty list!' % manifest)

    try:
        projects = d['projects']
    except KeyError:
        if len(includes) == 0:
            raise WSError('"projects" key missing in manifest %s' % manifest)
        else:
            return d

    for proj, props in projects.items():
        for prop in _REQUIRED_KEYS:
            if prop not in props:
                raise WSError('"%s" key missing from project %s in manifest'
                              % (prop, proj))

        for prop in props:
            if prop not in _ALL_KEYS:
                raise WSError('unknown key "%s" for project %s specified in '
                              'manifest' % (prop, proj))

    # Add computed keys.
    parent = os.path.realpath(os.path.join(root, os.pardir))
    for proj, props in projects.items():
        try:
            deps = props['deps']
        except KeyError:
            props['deps'] = ()
        else:
            if not isinstance(deps, list):
                raise WSError('"deps" key in project %s must be a list' % proj)
            if len(set(deps)) != len(deps):
                raise WSError('project %s has duplicate dependency' % proj)

        try:
            env = props['env']
        except KeyError:
            props['env'] = {}
        else:
            if not isinstance(env, dict):
                raise WSError('"env" key in project %s must be a dictionary'
                              % proj)
            for k, v in env.items():
                if not isinstance(k, str):
                    raise WSError('env key %s in project %s must be a string' %
                                  (k, proj))
                if not isinstance(v, str):
                    raise WSError('env value "%s" (key "%s") in project %s '
                                  'must be a string' % (v, k, proj))

        try:
            args = props['args']
        except KeyError:
            props['args'] = []
        else:
            if not isinstance(args, list):
                raise WSError('"args" key in project %s must be a list' % proj)
            for opt in args:
                if not isinstance(opt, str):
                    raise WSError('option "%s" in project %s must be a string'
                                  % (opt, proj))
            args = []
            for opt in props['args']:
                args.extend(opt.split())
            props['args'] = args

        try:
            args = props['targets']
        except KeyError:
            props['targets'] = DEFAULT_TARGETS
        else:
            if props['targets'] is None:
                props['targets'] = ()
            else:
                if not isinstance(props['targets'], list):
                    raise WSError('"targets" key in project %s must be a list'
                                  % proj)
                for target in props['targets']:
                    if not isinstance(target, str):
                        raise WSError('target "%s" in project %s must be a '
                                      'string' % (opt, proj))

        try:
            builder_args = props['builder-args']
        except KeyError:
            props['builder-args'] = {}
            builder_args = props['builder-args']
        else:
            if not isinstance(builder_args, dict):
                raise WSError('"builder-args" key in project %s must be a '
                              'list' % proj)
            for builder_arg in builder_args:
                if not isinstance(builder_arg, str):
                    raise WSError('builder arg "%s" in project %s must be '
                                  'a string' % (builder_arg, proj))

        props['path'] = os.path.join(parent, proj)

    return d


def include_paths(d, manifest):
    '''Return the manifest absolute paths included from the given parsed
    manifest.'''
    try:
        includes = d['include']
    except KeyError:
        return ()

    # The manifest's own directory is the default search path if none are
    # specified.
    search_paths = [os.path.realpath(os.path.join(manifest, os.pardir))]
    try:
        extra_search_paths = d['search-path']
    except KeyError:
        pass
    else:
        for path in extra_search_paths:
            path = os.path.realpath(os.path.join(manifest, os.pardir, path))
            search_paths.append(path)

    tweaked_includes = []
    for i, path in enumerate(includes):
        if path[0] == '/':
            # This is an absolute path, so no tweaking is necessary.
            tweaked_includes.append(path)
            continue

        # Relative path, so try to match it using the search paths.
        found_match = False
        for search_path in search_paths:
            full_path = os.path.realpath(os.path.join(search_path, path))
            if os.path.exists(full_path):
                found_match = True
                break
        if not found_match:
            raise WSError('cannot find manifest %s included by %s\n'
                          'search paths: %s' % (path, manifest, search_paths))

        # For directories, we include every file in the directory.
        if os.path.isdir(full_path):
            for filename in os.listdir(full_path):
                if not (filename.endswith('.yaml') or filename.endswith('.yml')):  # noqa: E501
                    continue
                filepath = os.path.join(full_path, filename)
                if not os.path.isfile(filepath):
                    continue
                tweaked_includes.append(filepath)
        else:
            tweaked_includes.append(full_path)

    return tweaked_includes


def merge_manifest(parent, parent_path, child, child_path):
    '''Merges the keys from the child dictionary into the parent dictionary. If
    there are conflicts, bail out with an error.'''
    try:
        parent_projects = parent['projects']
    except KeyError:
        parent_projects = {}
        parent['projects'] = parent_projects
    try:
        child_projects = child['projects']
    except KeyError:
        child_projects = {}
        child['projects'] = child_projects

    intersection = set(parent_projects).intersection(set(child_projects))
    if len(intersection) != 0:
        raise WSError('cannot include %s from %s, as the two share projects %s'
                      % (child_path, parent_path, intersection))
    parent_projects.update(child_projects)


def merge_includes(root, d, parent_manifest):
    '''Recursively merge all the include lines from the manifest into the given
    dictionary. Include paths are relative to the including manifest's parent
    directory.'''
    included = set(include_paths(d, parent_manifest))
    queue = collections.deque(included)
    while len(queue) > 0:
        manifest = queue.popleft()
        d_include = parse_yaml(root, manifest)
        log('merging manifest %s into %s' % (manifest, parent_manifest))
        merge_manifest(d, parent_manifest, d_include, manifest)
        for path in include_paths(d_include, manifest):
            # Prevent double-inclusion.
            if path in included:
                continue
            queue.append(path)
            included.add(path)


def parse_manifest_file(root, manifest):
    '''Parses the given ws manifest file, returning a dictionary of the
    manifest data.'''
    d = parse_yaml(root, manifest)
    merge_includes(root, d, manifest)

    # Compute reverse-dependency list.
    projects = d['projects']
    for proj, props in projects.items():
        props['downstream'] = []
    for proj, props in projects.items():
        deps = props['deps']
        for dep in deps:
            try:
                dep_props = projects[dep]
            except KeyError:
                raise WSError('project %s dependency %s not found in the '
                              'manifest' % (proj, dep))
            else:
                # Reverse-dependency list of downstream projects.
                dep_props['downstream'].append(proj)

    return projects


_WS_MANIFEST = None
def parse_manifest(root):  # noqa: E302
    '''Parses the ws manifest, returning a dictionary of the manifest data.'''
    # Parse.
    global _WS_MANIFEST
    if _WS_MANIFEST is None:
        _WS_MANIFEST = parse_manifest_file(root, get_manifest_link(root))
    return _WS_MANIFEST


def dependency_closure(d, projects):
    '''Returns the dependency closure for a list of projects. This is the set
    of dependencies of each project, dependencies of that project, and so
    on.'''
    # This set is for detecting circular dependencies.
    order = []
    processed = set()

    def process(project):
        processed.add(project)
        for dep in d[project]['deps']:
            if dep not in order:
                if dep in processed:
                    raise WSError('Projects %s and %s circularly depend on '
                                  'each other' % (project, dep))
                process(dep)
        order.append(project)

    for project in projects:
        if project not in processed:
            process(project)
    return tuple(order)


def find_root(start_dir):
    '''Recursively looks up in the directory hierarchy for a directory named
    .ws, and returns the first one found, or None if one was not found.'''
    path = os.path.realpath(start_dir)
    while path != '/':
        ws = os.path.join(path, '.ws')
        if os.path.isdir(ws):
            return ws
        path = os.path.realpath(os.path.join(path, os.pardir))
    return None


def get_default_manifest_name():
    '''Returns the name of the default ws manifest.'''
    return 'ws.yaml'


def get_ws_dir(root, ws):
    '''Returns the ws directory, given a directory obtained using
    find_root().'''
    return os.path.join(root, ws)


def get_ws_config_path(ws):
    '''Returns the ws internal config file, used for tracking the state of the
    workspace.'''
    return os.path.join(ws, 'config.yaml')


def get_ws_root(ws):
    '''Returns the root (.ws directory) given a workspace path.'''
    return os.path.realpath(os.path.join(ws, os.pardir))


def get_new_config(proj):
    '''Returns the dictionary representing the configuration for a newly-added
    project that has never been built before.'''
    return {'taint': False,
            'enable': True,
            'args': []
            }


_ORIG_WS_CONFIG = None
_WS_CONFIG = None
def get_ws_config(ws):  # noqa: E302
    '''Parses the current workspace config, returning a dictionary of the
    state.'''
    global _WS_CONFIG
    if _WS_CONFIG is None:
        config_path = get_ws_config_path(ws)
        with open(config_path, 'r') as f:
            _WS_CONFIG = yaml.safe_load(f)
            global _ORIG_WS_CONFIG
            # Save a copy of the config so we know later whether or not to
            # write it out when someone asks to sync the config.
            _ORIG_WS_CONFIG = copy.deepcopy(_WS_CONFIG)

    # Check if projects were added or removed from the manifest. If so, the
    # config needs to be updated accordingly.
    d = parse_manifest(get_ws_root(ws))
    deletions = []
    for proj in _WS_CONFIG['projects']:
        if proj not in d:
            deletions.append(proj)
    for proj in deletions:
        del _WS_CONFIG['projects'][proj]
        checksum_file = get_checksum_file(ws, proj)
        proj_dir = get_proj_dir(ws, proj)
        log('removing project %s, which is not in the manifest' % proj,
            logging.INFO)
        remove(checksum_file, True)
        rmtree(proj_dir, True)

    for proj in d:
        if proj in _WS_CONFIG['projects']:
            continue
        # Project is not in the config, so add it in.
        _WS_CONFIG['projects'][proj] = get_new_config(proj)

    # Split all project arguments by spaces so that args like '-D something'
    # turn into ['-D', 'something'], which is what exec requires. We do this
    # at parse time rather than in the "config" command to allow the user to
    # hand-edit in a natural way and have the config still work properly. The
    # next time the config is saved, it will be split by spaces.
    for proj, proj_map in _WS_CONFIG['projects'].items():
        parsed_args = []
        args = proj_map['args']
        for arg in args:
            parsed_args.extend(arg.split())
        proj_map['args'] = parsed_args

    return _WS_CONFIG


def _sync_file(f):
    '''Syncs a file to disk in whichever way we have available.'''
    # Not all systems have fdatasync. Empirically, the Mac appears not to have
    # it.
    if hasattr(os, 'fdatasync'):
        os.fdatasync(f)
    else:
        os.fsync(f)


def write_config(ws, config):
    '''Atomically updates the current ws config using the standard trick of
       writing to a tmp file in the same filesystem, syncing that file, and
       renaming it to replace the current contents.'''
    config_path = get_ws_config_path(ws)
    tmpfile = '%s.tmp' % config_path
    with open(tmpfile, 'w') as f:
        yaml.dump(config, f, default_flow_style=False)
        f.flush()
        _sync_file(f)
    os.rename(tmpfile, config_path)

    global _WS_CONFIG
    _WS_CONFIG = config


def sync_config(ws):
    '''Writes out the config if and only if it changed since we first read
       it.'''
    if _WS_CONFIG == _ORIG_WS_CONFIG:
        log('ws config did not change, so not updating')
        return
    log('updating config at %s' % ws)

    if dry_run():
        return

    write_config(ws, _WS_CONFIG)


def get_default_ws_name():
    '''Returns the name of the default workspace (the one you get when you
    don't explicitly give your workspace a name.'''
    return 'default'


def get_default_ws_link(root):
    '''Returns a path to the symlink that points to the current workspace.'''
    return os.path.join(root, get_default_ws_name())


def get_manifest_link_name():
    '''Returns the name of the symlink to the ws manifest.'''
    return 'manifest'


def get_manifest_link(root):
    '''Returns a path to the symlink that points to the ws manifest.'''
    return os.path.join(root, get_manifest_link_name())


def get_checksum_dir(ws):
    '''Returns the directory containing project build checksums.'''
    return os.path.join(ws, 'checksum')


def get_checksum_file(ws, proj):
    '''Returns the file containing the checksum for a given project.'''
    return os.path.join(get_checksum_dir(ws), proj)


def get_source_dir(root, d, proj):
    '''Returns the source code directory for a given project.'''
    parent = os.path.realpath(os.path.join(root, os.pardir))
    return os.path.join(parent, d[proj]['path'])


def get_toplevel_build_dir(ws):
    '''Returns the top-level directotory containing build artifacts for all
    projects.'''
    return os.path.join(ws, 'build')


def get_proj_dir(ws, proj):
    '''Returns the root directory for a given project.'''
    return os.path.join(get_toplevel_build_dir(ws), proj)


def get_source_link(ws, proj):
    '''Returns a path to the symlink inside the project directory that points
    back into the active source code (the code that the git-repo tool
    manages).'''
    return os.path.join(get_proj_dir(ws, proj), 'src')


def get_build_dir(ws, proj):
    '''Returns the build directory for a given project.'''
    return os.path.join(get_proj_dir(ws, proj), 'build')


def get_install_dir(ws, proj):
    '''Returns the install directory for a given project (the directory we use
    for --prefix and similar arguments.'''
    return os.path.join(get_build_dir(ws, proj), 'install')


_HOST_TRIPLET = None
def get_host_triplet():  # noqa: E302
    '''Gets the GCC host triplet for the current machine.'''
    global _HOST_TRIPLET
    if _HOST_TRIPLET is None:
        _HOST_TRIPLET = call_output(['gcc', '-dumpmachine'], override=True)
        _HOST_TRIPLET = _HOST_TRIPLET.rstrip()
    return _HOST_TRIPLET


def get_bin_paths(ws, proj):
    '''Gets the path to installed libraries for a project.'''
    noarch_bin_dir = os.path.join(get_install_dir(ws, proj), 'bin')
    arch_bin_arch_dir = os.path.join(noarch_bin_dir, get_host_triplet())
    return [noarch_bin_dir, arch_bin_arch_dir]


def get_lib_paths(ws, proj):
    '''Gets the path to installed libraries for a project.'''
    noarch_lib_dir = os.path.join(get_install_dir(ws, proj), 'lib')
    arch_lib_dir = os.path.join(noarch_lib_dir, get_host_triplet())
    return [noarch_lib_dir, arch_lib_dir]


def get_pkgconfig_paths(ws, proj):
    '''Gets the path to the .pc files for a project.'''
    pkgconfig_paths = []
    for lib_path in get_lib_paths(ws, proj):
        pkgconfig_paths.append(os.path.join(lib_path, 'pkgconfig'))
    return pkgconfig_paths


def set_stored_checksum(ws, proj, checksum):
    '''Sets the stored project checksum. This should be called after building
    the project.'''
    # Note that we don't worry about atomically writing this. The worst cases
    # are:
    # - We crash or get power loss and have a partial write. This causes a
    # corrupt checksum, which will not match the calculated checksum, be
    # stale, and cause us to redo the build. But an incremental build system
    # will cause this to have pretty low cost.
    # - The checksum is never updated when it should have been. Again, we'll
    # get a stale checksum and a rebuild, which shouldn't much hurt us.
    checksum_file = get_checksum_file(ws, proj)
    with open(checksum_file, 'w') as f:
        f.write('%s\n' % checksum)


def invalidate_checksum(ws, proj):
    '''Invalidates the current project checksum. This can be used to force a
    project to rebuild, for example if one of its dependencies rebuilds.'''
    log('invalidating checksum for %s' % proj)
    if dry_run():
        return

    try:
        remove(get_checksum_file(ws, proj))
    except OSError as e:
        if e.errno != errno.ENOENT:
            raise


def get_stored_checksum(ws, proj):
    '''Retrieves the currently stored checksum for a given project, or None if
    there is no checksum (either because the project was cleaned, or because
    we've never built the project before).'''
    if dry_run():
        return 'bogus-stored-checksum'

    checksum_file = get_checksum_file(ws, proj)
    try:
        with open(checksum_file, 'r') as f:
            checksum = f.read().rstrip()
    except IOError:
        return None

    # Note that we don't need to check if the checksum is corrupt. If it is, it
    # will not match the calculated checksum, so we will correctly see a stale
    # checksum.

    return checksum


def calculate_checksum(source_dir):
    '''Calculates and returns the SHA-1 checksum of a given git directory,
    including submodules and dirty files. This function should uniquely
    identify any source code that would impact the build. Note that we ignore
    files that have not been added to git but are in the git directory (files
    on which you have not run "git add". If this is not the case, it is likely
    a bug in the underlying project. Although we could use the find command
    instead of git, it is much slower and takes into account inconsequential
    files, like .cscope or .vim files that don't change the build (and that are
    typically put in .gitignore).

    It is very important that this function is both fast and accurate, as it is
    used to determine when projects need to be rebuilt, and thus gets run
    frequently. If this function gets too slow, working with ws will become
    painful. If this function is not accurate, then ws will have build bugs.'''
    # Collect the SHA-1 of HEAD and the diff of all dirty files.
    #
    # Note that it's very important to use the form "git diff HEAD" rather than
    # "git diff" or "git diff --cached" because "git diff HEAD" collects ALL
    # changes rather than just staged or unstaged changes.
    #
    # Additionally note the use of "submodule foreach --recursive", which will
    # recursively diff all submodules, submodules-inside-submodules, etc. This
    # ensures correctness even if deeply nested submodules change.
    head = call_git(source_dir, ('rev-parse', '--verify', 'HEAD'))
    repo_diff = call_git(source_dir,
                         ('diff',
                          'HEAD',
                          '--diff-algorithm=myers',
                          '--no-renames',
                          '--submodule=short'))
    submodule_diff = call_git(source_dir,
                              ('submodule',
                               'foreach',
                               '--recursive',
                               'git',
                               'diff',
                               'HEAD',
                               '--diff-algorithm=myers',
                               '--no-renames'))

    if dry_run():
        return 'bogus-calculated-checksum'

    # Finally, combine all data into one master hash.
    total = hashlib.sha1()
    total.update(head)
    total.update(repo_diff)
    total.update(submodule_diff)

    return total.hexdigest()


# These hooks contain functions to handle the build tasks for each build system
# we support. To add a new build system, add a new entry and supply the correct
# hooks.
_BUILD_TOOLS = {
    'meson': MesonBuilder,
    'cmake': CMakeBuilder,
    'setuptools': SetuptoolsBuilder
}


def get_builder(d, proj):
    '''Returns the build properties for a given project. This function should
    be used instead of directly referencing _BUILD_TOOLS.'''
    build = d[proj]['build']
    try:
        builder = _BUILD_TOOLS[build]
    except KeyError:
        raise WSError('unknown build tool %s for project %s'
                      % (build, proj))

    return builder


def merge_var(env, var, val):
    '''Merges the given value into the given environment variable using ":"
    syntax. This can be used to add an entry to LD_LIBRARY_PATH and other such
    environment variables.'''
    try:
        current = env[var]
    except KeyError:
        entries = val
    else:
        entries = current.split(':')
        if val not in entries:
            entries = val + entries
    env[var] = ':'.join(entries)


def expand_var(s, var, expansions):
    '''Expands a template string of the form ${VAR} with each entry of the
    expansions array, then join them in using the same scheme as the PATH
    variable. For example, given:
    s = ${ABC}/stuff
    var = ABC
    expansions = ['a', 'b', 'c'],
    then expand_sub would return 'a/stuff:b/stuff:c/stuff'.'''
    results = []
    for expansion in expansions:
        result = s.replace('${%s}' % var, expansion)
        # Don't add duplicate entries.
        if result not in results:
            results.append(result)
    return ':'.join(results)


def _merge_build_env(ws, d, proj, env, include_path):
    '''Merges the build environment for a single project into the given env.
    This is a helper function called by get_build_env for each dependency of a
    given project.'''
    build_dir = get_build_dir(ws, proj)
    bin_dirs = get_bin_paths(ws, proj)

    pkgconfig_paths = get_pkgconfig_paths(ws, proj)
    lib_paths = get_lib_paths(ws, proj)
    merge_var(env, 'PKG_CONFIG_PATH', pkgconfig_paths)
    merge_var(env, 'LD_LIBRARY_PATH', lib_paths)
    if include_path:
        merge_var(env, 'PATH', bin_dirs)

    # Add in any builder-specific environment tweaks.
    install_dir = get_install_dir(ws, proj)
    get_builder(d, proj).env(
        proj,
        install_dir,
        build_dir,
        env,
        d[proj]['builder-args'])

    # Add in any project-specific environment variables specified in the
    # manifest.
    for var, val in d[proj]['env'].items():
        val = expand_var(val, 'LIBDIR', lib_paths)
        val = expand_var(val, 'PREFIX', [install_dir])
        # Expand every environment variable.
        for k, v in env.items():
            val = expand_var(val, k, [v])
        merge_var(env, var, [val])


def get_build_env(ws, d, proj, include_path=False):
    '''Gets the environment that should be set during builds (and for the env
    command) for a given project.'''
    build_env = os.environ.copy()
    deps = dependency_closure(d, [proj])
    for dep in deps:
        _merge_build_env(ws, d, dep, build_env, include_path)

    return build_env



#!/usr/bin/python3
#
# Build action implementation.
#
# Copyright (c) 2018-2019 Xevo Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import errno
import logging
import multiprocessing
import os

from wst import (
    WSError,
    log
)
from wst.cmd import Command
from wst.cmd.clean import clean
from wst.conf import (
    calculate_checksum,
    dependency_closure,
    get_build_dir,
    get_build_env,
    get_builder,
    get_install_dir,
    get_proj_dir,
    get_source_dir,
    get_source_link,
    get_stored_checksum,
    get_ws_config,
    get_ws_dir,
    invalidate_checksum,
    parse_manifest,
    set_stored_checksum
)
from wst.shell import (
    rmtree,
    symlink,
    mkdir
)


def _build(root, ws, proj, d, current, ws_config, force):
    '''Builds a given project.'''
    if not ws_config['projects'][proj]['enable']:
        log('not building manually disabled project %s' % proj,
            logging.WARNING)
        return True

    if ws_config['projects'][proj]['taint']:
        log('force-cleaning tainted project %s' % proj, logging.WARNING)
        clean(root, ws, proj, d, True)

    source_dir = get_source_dir(root, d, proj)
    if not force:
        stored = get_stored_checksum(ws, proj)
        if current == stored:
            log('checksum for %s is current; skipping' % proj)
            return True
    else:
        log('forcing a build of %s' % proj)

    # Make the project directory if needed.
    proj_dir = get_proj_dir(ws, proj)
    try:
        mkdir(proj_dir)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise

    # Make the build directory if needed.
    build_dir = get_build_dir(ws, proj)
    try:
        mkdir(build_dir)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise
        needs_configure = False
    else:
        needs_configure = True

    # Populate the convenience source link.
    source_link = get_source_link(ws, proj)
    if not os.path.exists(source_link):
        source_dir = get_source_dir(root, d, proj)
        symlink(source_dir, source_link)

    # Invalidate the checksums for any downstream projects.
    for downstream_dep in d[proj]['downstream']:
        invalidate_checksum(ws, downstream_dep)

    # Add envs to find all projects on which this project is dependent.
    build_env = get_build_env(ws, d, proj)

    # Configure.
    builder = get_builder(d, proj)
    prefix = get_install_dir(ws, proj)
    extra_args = d[proj]['args'] + ws_config['projects'][proj]['args']
    if needs_configure:
        try:
            success = builder.conf(
                proj,
                prefix,
                source_dir,
                build_dir,
                build_env,
                ws_config['type'],
                d[proj]['builder-args'],
                extra_args)
        except Exception as _e:
            success = False
            e = _e
        else:
            e = None
        if not success:
            # Remove the build directory if we failed so that we are forced to
            # re-run configure next time.
            rmtree(build_dir)
            if e is not None:
                raise e
            else:
                return False

    # Build.
    success = builder.build(
        proj,
        prefix,
        source_dir,
        build_dir,
        build_env,
        d[proj]['targets'],
        d[proj]['builder-args'],
        d[proj]['args'])
    if success:
        set_stored_checksum(ws, proj, current)

    return success


class Build(Command):
    '''The build command.'''
    @classmethod
    def args(cls, parser):
        '''Populates the argument parser for the build command.'''
        parser.add_argument(
            'projects',
            action='store',
            nargs='*',
            help='Build a particular project or projects')
        parser.add_argument(
            '-f', '--force',
            action='store_true',
            default=False,
            help='Force a build')

    @classmethod
    def do(cls, ws, args):
        '''Executes the build subcmd.'''
        ws_config = get_ws_config(get_ws_dir(args.root, ws))
        d = parse_manifest(args.root)

        # Validate.
        for project in args.projects:
            if project not in d:
                raise WSError('unknown project %s' % project)

        if len(args.projects) == 0:
            projects = d.keys()
        else:
            projects = args.projects

        # Build in reverse-dependency order.
        order = dependency_closure(d, projects)

        # Get all checksums; since this is a nop build bottle-neck, do it in
        # parallel. On my machine, this produces a ~20% speedup on a nop
        # "build-all".
        pool = multiprocessing.Pool(multiprocessing.cpu_count())
        src_dirs = [get_source_dir(args.root, d, proj) for proj in order]
        checksums = pool.map(calculate_checksum, src_dirs)

        for i, proj in enumerate(order):
            logging.info('building %s' % proj)
            checksum = checksums[i]
            success = _build(
                args.root,
                ws,
                proj,
                d,
                checksum,
                ws_config,
                args.force)
            if not success:
                raise WSError('%s build failed' % proj)



#!/usr/bin/python3
#
# Rename action implementation.
#
# Copyright (c) 2018-2019 Xevo Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import os

from wst import WSError
from wst.cmd import Command
from wst.conf import (
    get_build_dir,
    get_default_ws_link,
    get_ws_dir,
    parse_manifest
)
from wst.shell import (
    remove,
    rename,
    symlink
)


class Rename(Command):
    '''The rename command.'''
    @classmethod
    def args(cls, parser):
        '''Populates the argument parser for the rename command.'''
        parser.add_argument(
            metavar='old-workspace-name',
            dest='old_ws',
            action='store',
            help='Old workspace name')
        parser.add_argument(
            metavar='new-workspace-name',
            dest='new_ws',
            action='store',
            help='New workspace name')

    @classmethod
    def do(cls, ws, args):
        '''Executes the rename command.'''
        if args.old_ws == 'default':
            raise WSError('cannot rename the default workspace; please use ws '
                          'default if you want to change it')

        old_ws_dir = get_ws_dir(args.root, args.old_ws)
        if not os.path.exists(old_ws_dir):
            raise WSError('workspace %s does not exist' % args.old_ws)

        d = parse_manifest(args.root)
        for proj in d:
            build_dir = get_build_dir(old_ws_dir, proj)
            if os.path.exists(build_dir):
                raise WSError('cannot rename a workspace that contains build '
                              'artifacts, as some builds contain absolute '
                              'paths and are thus not relocatable. Please '
                              'force-clean this workspace first and then '
                              'rename it.')

        new_ws_dir = get_ws_dir(args.root, args.new_ws)
        if os.path.exists(new_ws_dir):
            raise WSError('workspace %s already exists; please delete it '
                          'first if you want to do this rename' % args.new_ws)

        rename(old_ws_dir, new_ws_dir)
        default_link = get_default_ws_link(args.root)
        if os.readlink(default_link) == args.old_ws:
            remove(default_link)
            symlink(args.new_ws, default_link)



#!/usr/bin/python3
#
# Env action implementation.
#
# Copyright (c) 2018-2019 Xevo Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import argparse
import os

from wst import (
    WSError,
    log
)
from wst.cmd import Command
from wst.conf import (
    get_build_dir,
    get_build_env,
    merge_var,
    parse_manifest
)
from wst.shell import get_shell


class Env(Command):
    '''The env command.'''
    @classmethod
    def args(cls, parser):
        group = parser.add_mutually_exclusive_group()
        '''Populates the argument parser for the env command.'''
        group.add_argument(
            '-b', '--build-dir',
            action='store_true',
            default=False,
            help='Enter the env from the build directory instead of the '
                 'current directory')
        group.add_argument(
            '-c', '--current-dir',
            action='store',
            default=None,
            help='The directory from which the command will be run')

        parser.add_argument(
            'project',
            action='store',
            help='Enter the build environment for a particular project')
        parser.add_argument(
            'command',
            action='store',
            nargs=argparse.REMAINDER,
            help='Command to run inside the given environment')

    @classmethod
    def do(cls, ws, args):
        '''Executes the env command.'''
        build_dir = get_build_dir(ws, args.project)
        if not os.path.isdir(build_dir):
            raise WSError('build directory for %s doesn\'t exist; have you '
                          'built it yet?' % args.project)

        d = parse_manifest(args.root)
        build_env = get_build_env(ws, d, args.project, True)

        # Add the build directory to the path for convenience of running
        # non-installed binaries, such as unit tests.
        merge_var(build_env, 'PATH', [build_dir])

        if len(args.command) > 0:
            cmd = args.command
        else:
            cmd = [get_shell()]

        exe = os.path.basename(cmd[0])
        if exe == 'bash':
            # Tweak the prompt to make it obvious we're in a special env.
            prompt = ('\\[\033[1;32m\\][ws:%s env]\\[\033[m\\] \\u@\\h:\\w\\$ '  # noqa: E501
                      % args.project)
            build_env['PS1'] = prompt
            cmd.insert(1, '--norc')

        # Set an env var so the user can easily cd $WSBUILD and run tests or
        # similar inside the build directory.
        build_env['WSBUILD'] = build_dir

        log('execing with %s build environment: %s' % (args.project, cmd))

        if args.build_dir:
            args.current_dir = build_dir

        if args.current_dir is not None:
            os.chdir(args.current_dir)

        os.execvpe(cmd[0], cmd, build_env)



#!/usr/bin/python3
#
# setuptools builder.
#
# Copyright (c) 2018-2019 Xevo Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import os
from wst import (
    DEFAULT_TARGETS,
    WSError
)
from wst.builder import Builder
from wst.shell import (
    call_build,
    call_output,
    rmtree
)


def get_site_packages_dir(python_exe):
    '''Returns the site-packages directory for the given Python executable.'''
    return call_output(
        [python_exe,
         '-c',
         '\'import sysconfig;print(sysconfig.get_paths()[\"purelib\"])\''])


def get_python_exe(builder_args):
    '''Returns the Python executable to use.'''
    try:
        return builder_args['python-exe']
    except KeyError:
        import sys
        return sys.executable


class SetuptoolsBuilder(Builder):
    '''A setuptools builder.'''

    @classmethod
    def env(cls, proj, prefix, build_dir, env, builder_args):
        '''Sets up environment tweaks for setuptools.'''
        # Import here to prevent a circular import.
        from wst.conf import merge_var

        # setuptools won't install into a --prefix unless the site-packages
        # directory is in PYTHONPATH, so we'll add it in manually.
        python_path = os.path.join(
            prefix,
            get_site_packages_dir(get_python_exe(builder_args)))
        merge_var(env, 'PYTHONPATH', [python_path])

    @classmethod
    def conf(cls,
             proj,
             prefix,
             source_dir,
             build_dir,
             env,
             build_type,
             builder_args,
             args):
        '''Calls configure using setuptools.'''
        # setuptools doesn't have a configure step.
        return True

    @classmethod
    def build(cls,
              proj,
              prefix,
              source_dir,
              build_dir,
              env,
              targets,
              builder_args,
              args):
        '''Calls build using setuptools.'''
        if targets is not None and targets != DEFAULT_TARGETS:
            raise WSError('pip3 does not support alternate build targets but '
                          '"%s" was specified for targets' % targets)

        python_exe = get_python_exe(builder_args)
        cmd = [python_exe,
               '-m',
               'pip',
               'install',
               '--prefix=%s' % prefix,
               '--build=%s' % build_dir]
        cmd.extend(args)
        cmd.append('.')
        return call_build(cmd, cwd=source_dir, env=env)

    @classmethod
    def clean(cls, proj, prefix, source_dir, build_dir, env, builder_args):
        '''Calls clean using setuptools.'''
        # setuptools appears not to have a nicer way to do a "make clean" on a
        # dev-installed package.
        rmtree(build_dir)



#!/usr/bin/python3
#
# Config action implementation.
#
# Copyright (c) 2018-2019 Xevo Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import yaml

from wst import (
    BUILD_TYPES,
    log,
    WSError
)
from wst.cmd import Command
from wst.conf import (
    get_ws_config,
    parse_manifest
)


def parse_bool_val(val):
    '''Parses a value meant to be interpreted as a bool. Accepts 0, 1, false,
       and true (with any casing). Raises an exception if none match.'''
    if val is None:
        return True

    if val == '0':
        return False

    if val == '1':
        return True

    val = val.lower()
    if val == 'false':
        return False
    elif val == 'true':
        return True

    raise WSError('value "%s" is not a valid boolean' % val)


def _escape_commas(s):
    '''Turns "\\," into ",".'''
    return s.replace('\\,', ',')


def parse_build_args(val):
    '''Parses a value meant to be interpreted as comma-separated build
       arguments. \\, will escape a comma, so you can put a comma in a build
       argument.'''

    if val is None or val == '':
        return None

    split = []
    last_split = 0
    for i, c in enumerate(val):
        if c == ',':
            if not (i > 0 and val[i-1] == '\\'):
                substring = _escape_commas(val[last_split:i])
                split.append(substring)
                last_split = i+1
    split.append(_escape_commas(val[last_split:i+1]))
    return split


class Config(Command):
    '''The config command.'''
    @classmethod
    def args(cls, parser):
        '''Populates the argument parser for the config command.'''
        parser.add_argument(
            '-l', '--list',
            action='store_true',
            default=False,
            help='List the current workspace config')
        parser.add_argument(
            '-p', '--project',
            action='store',
            default=None,
            help='Set project-specific config')
        parser.add_argument(
            'options',
            action='store',
            nargs='*',
            help='Key-value options (format key=value')

    @classmethod
    def do(cls, ws, args):
        '''Executes the config command.'''
        config = get_ws_config(ws)
        if args.list:
            print(yaml.dump(config, default_flow_style=False), end='')
            return

        for arg in args.options:
            split = arg.split('=')
            split_len = len(split)
            key = split[0]
            if split_len == 1:
                val = None
            elif split_len == 2:
                val = split[1]
            else:
                val = '='.join(split[1:])

            project = args.project
            if project is not None:
                # Project-specific option.
                d = parse_manifest(args.root)
                if project not in d:
                    raise WSError('project "%s" not found' % project)

                can_taint = False
                if key == 'enable':
                    val = parse_bool_val(val)
                    can_taint = True
                elif key == 'args':
                    val = parse_build_args(val)
                    if val is None:
                        raise WSError('build args are not in the right format '
                                      '("args=key=val")')
                    can_taint = True
                else:
                    raise WSError('project key "%s" not found' % key)

                try:
                    proj_config = config['projects'][project]
                except KeyError:
                    proj_config = {}
                    config['projects'][project] = proj_config

                if can_taint and proj_config[key] != val:
                    log('tainting project %s' % project)
                    proj_config['taint'] = True

                proj_config[key] = val

            else:
                # Global option.
                if key == 'type':
                    if val is None or val not in BUILD_TYPES:
                        raise WSError('"type" key must be one of %s'
                                      % str(BUILD_TYPES))
                    if config[key] != val:
                        for proj_config in config['projects'].values():
                            proj_config['taint'] = True

                config[key] = val



#!/usr/bin/python3
#
# Init action implementation.
#
# Copyright (c) 2018-2019 Xevo Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import errno
import os

from wst import (
    BUILD_TYPES,
    WSError
)
from wst.cmd import Command
from wst.conf import (
    get_checksum_dir,
    get_default_ws_link,
    get_default_ws_name,
    get_default_manifest_name,
    get_manifest_link,
    get_manifest_link_name,
    get_new_config,
    get_toplevel_build_dir,
    get_ws_dir,
    parse_manifest_file,
    write_config
)
from wst.shell import (
    mkdir,
    symlink
)


MANIFEST_SOURCES = ('fs', 'repo')


def _is_subdir(a, b):
    '''Returns True if a is a sub-directory of b.'''
    a = os.path.abspath(a)
    b = os.path.abspath(b)

    return os.path.commonpath((a, b)) == b


class Init(Command):
    '''The list command.'''
    @classmethod
    def args(cls, parser):
        '''Populates the argument parser for the init subcmd.'''
        parser.add_argument(
            metavar='workspace',
            dest='init_ws',
            action='store',
            default=None,
            nargs='?',
            help='Workspace to initialize')

        parser.add_argument(
            '-t', '--type',
            action='store',
            choices=BUILD_TYPES,
            default='debug',
            help='Workspace type')
        parser.add_argument(
            '-m', '--manifest',
            action='store',
            default=get_default_manifest_name(),
            help='ws manifest path')
        parser.add_argument(
            '-s', '--manifest-source',
            action='store',
            choices=MANIFEST_SOURCES,
            required=True,
            help=('If -m is specified, what the path is relative to (see '
                  'README for details)'))

    @classmethod
    def do(cls, _, args):
        '''Executes the init command.'''
        if args.root is None:
            root = '.ws'
        else:
            root = args.root

        if args.init_ws is None:
            ws = 'ws'
        else:
            reserved = (get_default_ws_name(), get_manifest_link_name())
            for name in reserved:
                if args.init_ws == name:
                    raise WSError('%s is a reserved name; please choose a '
                                  'different name' % name)
            ws = args.init_ws
            if '.' in ws or '/' in ws:
                raise WSError('Workspace name "%s" contains an illegal '
                              'character (. or /). Please use a different '
                              'name.' % ws)
        ws_dir = get_ws_dir(root, ws)

        if os.path.exists(ws_dir):
            raise WSError('Cannot initialize already existing workspace %s'
                          % ws)

        if args.manifest_source == 'repo':
            parent = os.path.join(root, os.pardir)
            repo_manifest = os.path.realpath(os.path.join(parent,
                                                          '.repo',
                                                          'manifest.xml'))
            base = os.path.dirname(repo_manifest)
        elif args.manifest_source == 'fs':
            base = '.'
        else:
            raise NotImplementedError('Manifest source %s should be '
                                      'implemented'
                                      % args.manifest_source)
        manifest = os.path.join(base, args.manifest)
        if os.path.isdir(manifest):
            # If -m points to a directory instead of a file, assume there is a
            # file with the default manifest name inside.
            manifest = os.path.join(manifest, get_default_manifest_name())

        # Use a relative path for anything inside the parent of .ws and an
        # absolute path for anything outside. This is to maximize
        # relocatability for groups of git repos (e.g. using submodules,
        # repo-tool, etc.).
        manifest = os.path.abspath(manifest)
        parent = os.path.realpath(os.path.join(root, os.pardir))
        if _is_subdir(manifest, parent):
            manifest = os.path.relpath(manifest, root)
        else:
            manifest = os.path.abspath(manifest)

        # Make sure the given manifest is sane.
        if os.path.isabs(manifest):
            abs_manifest = manifest
        else:
            abs_manifest = os.path.abspath(os.path.join(root, manifest))
        d = parse_manifest_file(root, abs_manifest)

        try:
            mkdir(root)
            new_root = True
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise
            new_root = False

        try:
            mkdir(ws_dir)
            new_ws = True
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise
            new_ws = False

        if new_ws:
            # This is a brand-new workspace, so populate the initial workspace
            # directories.
            mkdir(get_toplevel_build_dir(ws_dir))
            mkdir(get_checksum_dir(ws_dir))

            proj_map = dict((proj, {}) for proj in d)
            for proj in proj_map:
                proj_map[proj] = get_new_config(proj)

            config = {
                'type': args.type,
                'projects': proj_map
            }

            write_config(ws_dir, config)

        if new_root:
            # This is a brand new root .ws directory, so populate the initial
            # symlink defaults.
            symlink(ws, get_default_ws_link(root))
            symlink(manifest, get_manifest_link(root))



#!/usr/bin/python3
#
# setuptools script for ws.
#
# Copyright (c) 2018-2019 Xevo Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import os
import setuptools
from wst.version import version


# Keep the long description in sync with the README.
script_dir = os.path.dirname(os.path.realpath(__file__))
readme_path = os.path.join(script_dir, 'README.md')
with open(readme_path, 'r') as f:
    long_desc = f.read()

setuptools.setup(
    name='wst',
    version=version(),
    description='A lightweight tool for managing a workspace of repositories',
    long_description=long_desc,
    long_description_content_type='text/markdown',
    author='Martin Kelly',
    author_email='mkelly@xevo.com',
    license='BSD',
    python_requires='>=3',
    install_requires=['PyYAML'],
    packages=setuptools.find_packages(),
    scripts=['bin/ws'],
    data_files=[('share/bash-completion/completions', ['bash-completion/ws'])],
    classifiers=['Development Status :: 3 - Alpha',
                 'Environment :: Console',
                 'Intended Audience :: Developers',
                 'License :: OSI Approved :: BSD License',
                 'Natural Language :: English',
                 'Operating System :: POSIX :: Linux',
                 'Programming Language :: Python :: 3 :: Only',
                 'Topic :: Software Development :: Build Tools',
                 ],
)
